 ; Copyright 2012-2016 Intel Corporation All Rights Reserved.
 ;
 ; The source code contained or described herein and all documents
 ; related to the source code ("Material") are owned by Intel
 ; Corporation or its suppliers or licensors. Title to the Material
 ; remains with Intel Corporation or its suppliers and licensors. The
 ; Material may contain trade secrets and proprietary and confidential
 ; information of Intel Corporation and its suppliers and licensors,
 ; and is protected by worldwide copyright and trade secret laws and
 ; treaty provisions. No part of the Material may be used, copied,
 ; reproduced, modified, published, uploaded, posted, transmitted,
 ; distributed, or disclosed in any way without Intel's prior express
 ; written permission.
 ;
 ; No license under any patent, copyright, trade secret or other
 ; intellectual property right is granted to or conferred upon you by
 ; disclosure or delivery of the Materials, either expressly, by
 ; implication, inducement, estoppel or otherwise. Any license under
 ; such intellectual property rights must be express and approved by
 ; Intel in writing.
 ;
 ; Unless otherwise agreed by Intel in writing, you may not remove or
 ; alter this notice or any other notice embedded in Materials by Intel
 ; or Intel's suppliers or licensors in any way.
(defclass _STATISTICS
  "Class to calculate statistics for a 1 dimensional vector of
   numbers (integer and/or floating point).  Currently supported
   statistics are:

     - count: number of data elements
     - mad: the median absolute deviation
     - mean: the average value
     - median: the median value
     - stddev: the standard deviation
     - sum: the sum of all values

   Once a statistical value has been calculated, it is cached to
   improve performance.  As long as a new data element is not added,
   the cached value will be used."
  (is-a USER)
  (role concrete)
  (pattern-match non-reactive)

  (slot cache-mad (type SYMBOL)
        (allowed-values TRUE FALSE)
        (default FALSE)
        (visibility private))
  (slot cache-mean (type SYMBOL)
        (allowed-values TRUE FALSE)
        (default FALSE)
        (visibility private))
  (slot cache-median (type SYMBOL)
        (allowed-values TRUE FALSE)
        (default FALSE)
        (visibility private))
  (slot cache-stddev (type SYMBOL)
        (allowed-values TRUE FALSE)
        (default FALSE)
        (visibility private))
  (slot cache-sum (type SYMBOL)
        (allowed-values TRUE FALSE)
        (default FALSE)
        (visibility private))
  (multislot data (type NUMBER)
             (visibility private))
  (slot mad (type NUMBER)
        (visibility private))
  (slot mean (type NUMBER)
        (visibility private))
  (slot median (type NUMBER)
        (visibility private))
  (slot stddev (type NUMBER)
        (visibility private))
  (slot sum (type NUMBER)
        (visibility private))
  (slot outlier-margin (type NUMBER)
        (visibility private)))

(defmessage-handler _STATISTICS _median (?list)
  "Helper to calculate the median of a sorted list of numbers.  The
   list must already be sorted!"
  (bind ?count (length$ ?list))
  (if (oddp ?count) then ; odd, use the center element
    (bind ?index (/ (+ ?count 1) 2))
    (return (nth$ ?index ?list))
    else ; even, average of the two center elements
    (bind ?index1 (/ ?count 2))
    (bind ?index2 (+ ?index1 1))
    (return (/ (+ (nth$ ?index1 ?list) (nth$ ?index2 ?list)) 2))))

(defmessage-handler _STATISTICS add (?v)
  "Add a new data element to the front of the list and invalidate any
   cached values."
  (bind ?self:cache-mad FALSE)
  (bind ?self:cache-mean FALSE)
  (bind ?self:cache-median FALSE)
  (bind ?self:cache-stddev FALSE)
  (bind ?self:cache-sum FALSE)
  (slot-insert$ ?self data 1 ?v))

(defmessage-handler _STATISTICS count ()
  "Return the number of data elements"
  (return (length$ ?self:data)))

(defmessage-handler _STATISTICS mad ()
  "Return the median absolute deviation"
  (if ?self:cache-mad then (return ?self:mad))
  (bind ?median (send ?self median))
  (bind ?list (create$ ))
  (progn$ (?d ?self:data)
          (bind ?list (insert$ ?list 1 (abs (- ?d ?median)))))
  (bind ?list (sort > ?list))
  (bind ?self:mad (send ?self _median ?list))
  (bind ?self:cache-mad TRUE)
  (return ?self:mad))

(defmessage-handler _STATISTICS mean ()
  "Return the arithmetic mean value"
  (if ?self:cache-mean then (return ?self:mean))
  (bind ?self:mean (/ (send ?self sum) (send ?self count)))
  (bind ?self:cache-mean TRUE)
  (return ?self:mean))

(defmessage-handler _STATISTICS median ()
  "Return the median value"
  (if ?self:cache-median then (return ?self:median))
  (bind ?self:data (sort > ?self:data))
  (bind ?self:median (send ?self _median ?self:data))
  (bind ?self:cache-median TRUE)
  (return ?self:median))

(defmessage-handler _STATISTICS stddev ()
  "Return the standard deviation value.  The standard deviation is not
   defined for a single data element."
  (if ?self:cache-stddev then (return ?self:stddev))
  (if (< (length$ ?self:data) 2) ; need at least 2 elements
      then (return FALSE))
  (bind ?mean (send ?self mean))
  (bind ?sum 0)
  (progn$ (?d ?self:data)
          (bind ?sum (+ ?sum (** (- ?d ?mean) 2))))
  (bind ?self:stddev (sqrt (/ ?sum (- (send ?self count) 1))))
  (bind ?self:cache-stddev TRUE)
  (return ?self:stddev))

(defmessage-handler _STATISTICS sum ()
  "Return the sum"
  (if ?self:cache-sum then (return ?self:sum))
  (bind ?sum 0)
  (progn$ (?d ?self:data)
          (bind ?sum (+ ?sum ?d)))
  (bind ?self:sum ?sum)
  (bind ?self:cache-sum TRUE)
  (return ?self:sum))


(defclass STATISTICS
  "Inheriting this class provides a class with access to statistical
   capabilities.  The slots in the class are shared so the slots are
   globally accessible from all inheriting classes.

   The class is a key / value map.  Keys may be arbitrarily defined
   and map to instances of the _STATISTICS class.  Any number of keys
   may be created per class, but keys are global, so it's highly
   recommended that the key include the name of the class.  If per
   node statistics are needed, the key should also include the
   node_id.  Further, if statistics are needed for multiple slots in
   the same class, include the slot name in the key.  For example, the
   IMB_PINGPONG class keys might be IMB_PINGPONG-node00-latency and
   IMB_PINGPONG-node00-bandwidth.

   One-to-one wrapper message handlers are provided for the
   _STATISTICS message handlers.

   To gain statistical capabilities, a class should do the following:

     - Inherit STATISTICS
       E.g.,
       (defclass FOO (is-a USER STATISTICS)
         ...
         (slot bar (type NUMBER))
         ...

     - Wrap the put message handler for each of the class slots
       wanting statistical capabilities to add the value using one or
       more keys.
       E.g.,

       (defmessage-handler FOO put-bar after (?v)
         (bind ?key1 (sym-cat (class ?self) - ?self:node_id))
         (send ?self add ?key1 ?v)
         (bind ?key2 (class ?self))
         (send ?self add ?key2 ?v))

       Alternatively, the init message handler could be wrapped:

       (defmessage-handler FOO init after ()
         (bind ?key1 (sym-cat (class ?self) - ?self:node_id))
         (send ?self add ?key1 ?self:bar)
         (bind ?key2 (class ?self))
         (send ?self add ?key2 ?self:bar))

   Note: modifying or removing data elements once they have been
   added is not supported."
  (is-a USER)
  (role abstract)

  (slot statistics_control (default FALSE) (storage shared))
  (multislot _statistics_keys (type SYMBOL)
             (pattern-match non-reactive) ; critical for performance
             (storage shared)
             (visibility private))
  (multislot _statistics_instances (type INSTANCE)
             (allowed-classes _STATISTICS)
             (pattern-match non-reactive) ; critical for performance
             (storage shared)
             (visibility private)))

(defrule flip_statistics_control
  "Rules start matching immediately even though additional data points
   may have yet to be created.  To force rules to delay until all the
   instances have been defined, use a control variable.  This rule
   will always activate due to the empty left hand side, but the
   control variable will not be changed to the TRUE state until the
   run command is issued.  Since the slot is shared, flipping the
   control variable in any one instance will do."
  =>
  (do-for-instance ((?i STATISTICS)) TRUE
                   (send ?i put-statistics_control TRUE)))

(defmessage-handler STATISTICS _find_pair (?key)
  "Helper to look up the _STATISTICS instance corresponding to the
   key."
  (bind ?pos (member$ ?key ?self:_statistics_keys))
  (if ?pos
      then (return (nth$ ?pos ?self:_statistics_instances))
      else (return FALSE)))

(defmessage-handler STATISTICS add (?key ?v)
  "Add the value to corresponding _STATISTICS instance.  If the key /
   value pair does not already exist, create it."
  (bind ?pos (member$ ?key ?self:_statistics_keys))
  (if ?pos then
    ; key/instance pair already exists
    (bind ?instance (nth$ ?pos ?self:_statistics_instances))
    else
    ; key/instance pair needs to be created and added
    (bind ?instance (make-instance of _STATISTICS))
    (slot-insert$ ?self _statistics_keys 1 ?key)
    (slot-insert$ ?self _statistics_instances 1 ?instance))
  (send ?instance add ?v))

(defmessage-handler STATISTICS count (?key)
  "Wrap the corresponding _STATISTICS message handler"
  (bind ?instance (send ?self _find_pair ?key))
  (if ?instance then (send ?instance count)))

(defmessage-handler STATISTICS mad (?key)
  "Wrap the corresponding _STATISTICS message handler"
  (bind ?instance (send ?self _find_pair ?key))
  (if ?instance then (send ?instance mad)))

(defmessage-handler STATISTICS mean (?key)
  "Wrap the corresponding _STATISTICS message handler"
  (bind ?instance (send ?self _find_pair ?key))
  (if ?instance then (send ?instance mean)))

(defmessage-handler STATISTICS median (?key)
  "Wrap the corresponding _STATISTICS message handler"
  "Documentation"
  (bind ?instance (send ?self _find_pair ?key))
  (if ?instance then (send ?instance median)))

(defmessage-handler STATISTICS stddev (?key)
  "Wrap the corresponding _STATISTICS message handler"
  (bind ?instance (send ?self _find_pair ?key))
  (if ?instance then (send ?instance stddev)))

(defmessage-handler STATISTICS sum (?key)
  "Wrap the corresponding _STATISTICS message handler"
  (bind ?instance (send ?self _find_pair ?key))
  (if ?instance then (send ?instance sum)))

(defmessage-handler STATISTICS outlier-margin (?key)
  (bind ?instance (send ?self _find_pair ?key))
  (if ?instance then (send ?instance sum)))


(defmessage-handler STATISTICS mad_too_small (?key ?max-median-mad-dist)
  "determines if the log10 distance between the median and mad is greater
   than the max distance."
  (bind ?instance (send ?self _find_pair ?key))

  ;get the absolute values of the median and mad
  (bind ?median (send ?instance median))
  (bind ?absmedian (abs ?median))

  (bind ?mad (send ?instance mad))
  (bind ?absmad (abs ?mad))

  ;get the log10 values
  (if (or (eq ?absmedian 0) (eq ?absmedian 0.0))
      then (bind ?log-median ?absmedian)
      else (bind ?log-median (log10 ?absmedian)))

  (if (or (eq ?absmad 0) (eq ?absmad 0.0)) then (bind ?log-mad ?absmad)
      else (bind ?log-mad (log10 ?absmad)))

  ;compare the distance to the threshold
  (if (> (- ?log-median ?log-mad) ?max-median-mad-dist) then (return TRUE))

  (return FALSE)
)


(defmessage-handler STATISTICS outlier (?key ?val ?num-mad ?max-median-mad-dist
                                             ?median-pct)
  "Calculate if value is an outlier from the list at the key using the
   median absolute deviation, if the MAD is too small, then use a percentage
   of the median"

  (bind ?instance (send ?self _find_pair ?key))
  (bind ?median (send ?instance median))

  ; if the mad is too small, then set the margin to be a percentage of the
  ; median
  (if (send ?self mad_too_small ?key ?max-median-mad-dist)
      then (bind ?margin (* ?median (/ ?median-pct 100.0)))

      ; else set the margin to be the MAD
      else (bind ?mad (send ?instance mad))
           (bind ?margin (* ?mad ?num-mad)))

  ;check for outlier
  (if (or (> ?val (+ ?median ?margin))
          (< ?val (- ?median ?margin)))
      then
          ; if the sample is an outlier, then set margin slot in the statistics
          ; instance
          (send ?instance put-outlier-margin ?margin)
          (return TRUE))

  (return FALSE)
)


(defmessage-handler STATISTICS outlier-severity (?key ?val ?min ?max)
  "Calculate the severity using the normal distribution.
   TODO: Replace the gaussian probability density function with an
   exponential distribution that has a negative skew.
   Also potentially add arguments for the desired distribution and moments."

  (bind ?instance (send ?self _find_pair ?key))
  (bind ?median (send ?instance median))
  (bind ?margin (send ?instance get-outlier-margin))

  ; if the margin is zero, then the severity is the minimum
  (if (or (= ?margin 0) (= ?margin 0.0)) then (bind ?pt 1.0)
      else
      ; apply the gaussian pdf to the value
      (bind ?pt (- 1 (exp (/ (* (- ?val ?median) (- ?median ?val))
                             (* 2 ?margin ?margin))))))

  ;scale the point with the range of severities
  (return (+ ?min (* (- ?max ?min) ?pt))))
